# Copyright 2021 Vincent Texier <vit@free.fr>
# Copyright 2023 Greenbone AG
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import sys
from typing import Iterable, List, Optional, Tuple

from autohooks.api import error, ok, out
from autohooks.api.git import get_staged_status, stash_unstaged_changes
from autohooks.api.path import match
from autohooks.config import Config

DEFAULT_INCLUDE: Tuple[str] = ("*.py",)
DEFAULT_ARGUMENTS: List[str] = []


def check_mypy_installed() -> None:
    try:
        import mypy  # pylint: disable=import-outside-toplevel, disable=unused-import
    except ImportError as e:
        raise RuntimeError(
            "Could not find mypy. Please add mypy to your python environment"
        ) from e


def get_mypy_config(config: Config) -> Config:
    return config.get("tool").get("autohooks").get("plugins").get("mypy")


def ensure_iterable(value) -> Iterable[str]:
    if isinstance(value, str):
        return [value]

    return value


def get_include_from_config(config: Optional[Config]) -> Iterable[str]:
    if not config:
        return DEFAULT_INCLUDE

    mypy_config = get_mypy_config(config)
    include = ensure_iterable(mypy_config.get_value("include", DEFAULT_INCLUDE))

    return include


def get_mypy_arguments(config: Optional[Config]) -> Iterable[str]:
    if not config:
        return DEFAULT_ARGUMENTS

    mypy_config = get_mypy_config(config)
    arguments = ensure_iterable(
        mypy_config.get_value("arguments", DEFAULT_ARGUMENTS)
    )

    return arguments


# pylint: disable=unused-argument
def precommit(config: Optional[Config] = None, **kwargs) -> int:
    check_mypy_installed()
    include = get_include_from_config(config)

    files = [f for f in get_staged_status() if match(f.path, include)]

    if not files:
        ok("No staged files to check.")
        return 0

    arguments = get_mypy_arguments(config)

    with stash_unstaged_changes(files):
        ret = 0

        absolute_path_files = list(map(lambda p: str(p.absolute_path()), files))

        cmd = ["mypy"]
        cmd.extend(arguments)
        cmd.extend(absolute_path_files)
        try:
            subprocess.run(cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            ret = e.returncode
            error("Static typing error(s) found:")
            lint_errors = e.stdout.decode(
                encoding=sys.getdefaultencoding(), errors="replace"
            ).split("\n")
            # Skip the first line that only shows ******** Module blah
            for line in lint_errors:
                out(line)
            return ret

        ok("Mypy check was successful.")
        return ret



#!/usr/bin/env python3

# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from autohooks.api.git import StatusEntry
from autohooks.config import load_config_from_pyproject_toml

from autohooks.plugins.mypy.mypy import (
    DEFAULT_ARGUMENTS,
    DEFAULT_INCLUDE,
    check_mypy_installed,
    ensure_iterable,
    get_include_from_config,
    get_mypy_arguments,
    get_mypy_config,
    precommit,
)


def get_test_config_path(name):
    return Path(__file__).parent / name


class AutohooksMypyTestCase(TestCase):
    def test_mypy_installed(self):
        path = sys.path
        sys.path = []
        with self.assertRaises(Exception):
            check_mypy_installed()
        sys.path = path

    def test_get_mypy_arguments(self):
        args = get_mypy_arguments(config=None)
        self.assertEqual(args, DEFAULT_ARGUMENTS)

    def test_get_mypy_config(self):
        config_path = get_test_config_path("pyproject.test.toml")
        self.assertTrue(config_path.is_file())

        autohooksconfig = load_config_from_pyproject_toml(config_path)

        mypy_config = get_mypy_config(autohooksconfig.get_config())
        self.assertEqual(mypy_config.get_value("foo"), "bar")

    def test_ensure_iterable(self):
        foo = "bar"  # pylint: disable=disallowed-name
        bar = ensure_iterable(foo)  # pylint: disable=disallowed-name
        self.assertEqual(bar, ["bar"])

        foo = ["bar"]  # pylint: disable=disallowed-name
        bar = ensure_iterable(foo)  # pylint: disable=disallowed-name
        self.assertEqual(bar, ["bar"])

    def test_get_include_from_config(self):
        include = get_include_from_config(config=None)
        self.assertEqual(include, DEFAULT_INCLUDE)

    @patch("autohooks.plugins.mypy.mypy.ok")
    def test_precommit_no_files(self, _ok_mock):
        ret = precommit()
        self.assertFalse(ret)

    # these Terminal output functions don't run in the CI ...
    # @patch('sys.stdout', new_callable=StringIO)
    @patch("autohooks.plugins.mypy.mypy.ok")
    @patch("autohooks.plugins.mypy.mypy.out")
    @patch("autohooks.plugins.mypy.mypy.error")
    @patch("autohooks.plugins.mypy.mypy.get_staged_status")
    def test_precommit_errors(
        self,
        staged_mock,
        _error_mock,
        _out_mock,
        _ok_mock,  # _mock_stdout
    ):
        code = """from typing import List

bad_type: List[str] = 'tmp.txt'
"""

        test_file = Path(__file__).parent / "static_typing_test.py"
        with open(test_file, "w", encoding="utf-8") as fp:
            fp.writelines(code)

        staged_mock.return_value = [
            StatusEntry(
                status_string="M  static_typing_test.py",
                root_path=Path(__file__).parent,
            )
        ]

        ret = precommit()

        # Returncode != 0 -> errors
        self.assertEqual(1, ret)
        test_file.unlink()

    # these Terminal output functions don't run in the CI ...
    # @patch('sys.stdout', new_callable=StringIO)
    @patch("autohooks.plugins.mypy.mypy.ok")
    @patch("autohooks.plugins.mypy.mypy.out")
    @patch("autohooks.plugins.mypy.mypy.error")
    @patch("autohooks.plugins.mypy.mypy.get_staged_status")
    def test_precommit_ok(
        self,
        staged_mock,
        _error_mock,
        _out_mock,
        _ok_mock,  # _mock_stdout
    ):
        code = """from typing import List

bad_type: List[str] = ['tmp.txt']
"""

        test_file = Path(__file__).parent / "static_typing_test.py"
        with open(test_file, "w", encoding="utf-8") as fp:
            fp.writelines(code)

        staged_mock.return_value = [
            StatusEntry(
                status_string="M  static_typing_test.py",
                root_path=Path(__file__).parent,
            )
        ]

        ret = precommit()

        # Returncode 0 -> no errors
        self.assertEqual(0, ret)
        test_file.unlink()
